```@meta
CurrentModule = Oscar
DocTestSetup = quote
  using Oscar
end
```

```@setup oscar
using Oscar
```

```@contents
Pages = ["binomial_ideals.md"]
```

# Binomial Primary Decomposition

## Introduction

A *binomial* is a polynomial consisting of at most two terms. A *binomial ideal*
is an ideal which can be generated by binomials. A *binomial primary decomposition*
is a primary decomposition of a binomial ideal into primary ideals which are binomial
as well. In this section, focusing on polynomial rings over fields, we discuss functionality
for computing such decompositions.

We begin by recalling that a proper ideal $I$ of a polynomial ring $K[x_1, \dots, x_n]$ over a
field $K$ is called *cellular* if each variable $x_i$ is either a nonzerodivisor or nilpotent modulo $I$.
In this case, we refer to the nonzerodivisors among the variables as the *cell variables* with respect to $I$.
A *cellular decomposition* of a proper binomial ideal $I$ of $K[x_1, \dots, x_n]$ is a decomposition of $I$ into cellular binomial ideals.
Using Noetherian induction, it is not too difficult to show that such decompositions exist. 

With this notation, the algorithms for computing binomial primary decompositions proceed in two main steps:

- First, compute a cellular decomposition of the given binomial ideal.
- Then, decompose each cellular component into primary binomial ideals.

While the first step can be performed over any ground field for which Gröbner bases are implemented, the second step may require a field extension: From a theoretical point of view, the existence of binomial primary decompositions is only guaranteed if the ground field is algebraically closed.

!!! note
    A *pure difference binomial* is a binomial which is the difference of two monomials. A *unital binomial ideal* is an ideal which can be generated by pure difference binomials and monomials. Note that cellular components of unital binomial ideals are unital as well. For unital binomial ideals in $\mathbb Q[x_1, \dots, x_n]$, binomial primary decompositions exist already over cyclotomic extensions of $\mathbb Q$. In particular, any such ideal can be decomposed over the abelian closure $\mathbb Q^{\text{ab}}$ of $\mathbb Q$. While OSCAR offers functionality for doing this, computing  binomial primary decompositions  in other cases is not yet supported. 

    See the number theory chapter for how to deal with $\mathbb Q^{\text{ab}}$.
!!! note
    Binomial primary decompositions computed with OSCAR are not necessarily minimal.

Papers offering details on theory and algorithms as well as examples include:

- [ES96](@cite)
- [OMdCS00](@cite)
- [Kah10](@cite)
- [EM16](@cite)
- [EM19](@cite)

## Basic Tests

### Binomiality Test

```@docs
is_binomial(f::MPolyRingElem)
is_binomial(I::MPolyIdeal)
```
```jldoctest
julia> R, (x, y, z) = polynomial_ring(QQ, ["x", "y", "z"])
(Multivariate Polynomial Ring in x, y, z over Rational Field, QQMPolyRingElem[x, y, z])

julia> f = 2*x+y
2*x + y

julia> is_binomial(f)
true

julia> J = ideal(R, [x^2-y^3, z^2])
ideal(x^2 - y^3, z^2)

julia> is_binomial(J)
true

```

### Cellularity Test

```@docs
is_cellular(I::MPolyIdeal)
```

### Unitality Test

```@docs
is_unital(I::MPolyIdeal)
```



## Cellular Decomposition

```@docs
cellular_decomposition(I::MPolyIdeal)
```

## Primary Decomposition of Cellular Ideals

```@docs
cellular_hull(I::MPolyIdeal{QQMPolyRingElem})
cellular_associated_primes(I::MPolyIdeal{QQMPolyRingElem})
cellular_minimal_associated_primes(I::MPolyIdeal{QQMPolyRingElem})
cellular_primary_decomposition(I::MPolyIdeal{QQMPolyRingElem})
```

## Primary Decomposition of Binomial  Ideals

```@docs
binomial_primary_decomposition(I::MPolyIdeal{QQMPolyRingElem})
```


