
########################################################################
# Attributes of AbsCoveredScheme                                       #
########################################################################

########################################################################
# Type getters                                                         #
########################################################################
base_ring_type(::Type{T}) where {BRT, T<:AbsCoveredScheme{BRT}} = BRT
base_ring_type(X::AbsCoveredScheme) = base_ring_type(typeof(X))

########################################################################
# Basic getters                                                        #
########################################################################
base_ring(X::AbsCoveredScheme) = base_ring(underlying_scheme(X))

@doc raw"""
    coverings(X::AbsCoveredScheme)

Return the list of internally stored `Covering`s of ``X``.

# Examples
```jldoctest
julia> P = projective_space(QQ, 2);

julia> Pcov = covered_scheme(P);

julia> coverings(Pcov)
1-element Vector{Covering{QQField}}:
 Covering with 3 patches
```
"""
function coverings(X::AbsCoveredScheme) ::Vector{<:Covering}
  return coverings(underlying_scheme(X))
end

@doc raw"""
    default_covering(X::AbsCoveredScheme)

Return the default covering for ``X``.

# Examples
```jldoctest
julia> P = projective_space(QQ, 2);

julia> S = homogeneous_coordinate_ring(P);

julia> I = ideal(S, [S[1]*S[2]-S[3]^2]);

julia> X = subscheme(P, I);

julia> Xcov = covered_scheme(X)
covered scheme with 3 affine patches in its default covering

julia> default_covering(Xcov)
Covering with 3 patches

```
"""
function default_covering(X::AbsCoveredScheme)
  return default_covering(underlying_scheme(X))::Covering
end

@doc raw"""
    patches(X::AbsCoveredScheme) = patches(default_covering(X))

Return the affine patches in the `default_covering` of ``X``.
"""
patches(X::AbsCoveredScheme) = patches(default_covering(X))

@doc raw"""
    affine_charts(X::AbsCoveredScheme)

Return the affine charts in the `default_covering` of ``X``.

# Examples
```jldoctest
julia> P = projective_space(QQ, 2);

julia> S = homogeneous_coordinate_ring(P);

julia> I = ideal(S, [S[1]*S[2]-S[3]^2]);

julia> X = subscheme(P, I);

julia> Xcov = covered_scheme(X)
covered scheme with 3 affine patches in its default covering

julia> affine_charts(Xcov)
3-element Vector{AbsSpec}:
 Spec of Quotient of Multivariate Polynomial Ring in (s1//s0), (s2//s0) over Rational Field by ideal((s1//s0) - (s2//s0)^2)
 Spec of Quotient of Multivariate Polynomial Ring in (s0//s1), (s2//s1) over Rational Field by ideal((s0//s1) - (s2//s1)^2)
 Spec of Quotient of Multivariate Polynomial Ring in (s0//s2), (s1//s2) over Rational Field by ideal((s0//s2)*(s1//s2) - 1)

```
"""
affine_charts(X::AbsCoveredScheme) = basic_patches(default_covering(X))


########################################################################
# Attributes of CoveredScheme                                          #
########################################################################

########################################################################
# Basic getters                                                        #
########################################################################
coverings(X::CoveredScheme) = X.coverings
default_covering(X::CoveredScheme) = X.default_covering
patches(X::CoveredScheme) = patches(default_covering(X))
glueings(X::CoveredScheme) = glueings(default_covering(X))
base_ring(X::CoveredScheme) = X.kk


########################################################################
# Names of CoveredSchemes                                              #
########################################################################
set_name!(X::AbsCoveredScheme, name::String) = set_attribute!(X, :name, name)
name(X::AbsCoveredScheme) = get_attribute(X, :name)::String
has_name(X::AbsCoveredScheme) = has_attribute(X, :name)

########################################################################
# Auxiliary attributes                                                 #
########################################################################
function dim(X::AbsCoveredScheme) 
  if !has_attribute(X, :dim)
    d = -1
    is_equidimensional=true
    for U in patches(default_covering(X))
      e = dim(U)
      if e > d
        d == -1 || (is_equidimensional=false)
        d = e
      end
    end
    set_attribute!(X, :dim, d)
    if !is_equidimensional
      # the above is not an honest check for equidimensionality,
      # because in each chart the output of `dim` is only the 
      # supremum of all components. Thus we can only infer 
      # non-equidimensionality in case this is already visible
      # from comparing the different charts
      set_attribute!(X, :is_equidimensional, false)
    end
  end
  return get_attribute(X, :dim)::Int
end

@attr function singular_locus_reduced(X::AbsCoveredScheme)
  D = IdDict{AbsSpec, Ideal}()
  for U in affine_charts(X)
    _, inc_sing = singular_locus_reduced(U)
    D[U] = image_ideal(inc_sing)
  end
  Ising = IdealSheaf(X, D)
  inc = CoveredClosedEmbedding(X, Ising)
  return domain(inc), inc
end

@attr function singular_locus(
    X::AbsCoveredScheme;
  )
  D = IdDict{AbsSpec, Ideal}()
  covering = (has_attribute(X, :simplified_covering) ? simplified_covering(X) : default_covering(X))
  for U in covering
    _, inc_sing = singular_locus(U)
    D[U] = image_ideal(inc_sing)
  end
  Ising = IdealSheaf(X, D)
  inc = CoveredClosedEmbedding(X, Ising)
  return domain(inc), inc
end

@attr function ideal_sheaf_of_singular_locus(
    X::AbsCoveredScheme;
  )
  D = IdDict{AbsSpec, Ideal}()
  covering = (has_attribute(X, :simplified_covering) ? simplified_covering(X) : default_covering(X))
  for U in covering
    _, inc_sing = singular_locus(U)
    D[U] = radical(image_ideal(inc_sing))
  end
  Ising = IdealSheaf(X, D, check=false)
  return Ising
end

function simplified_covering(X::AbsCoveredScheme)
  if !has_attribute(X, :simplified_covering)
    simplify!(X)
  end
  return get_attribute(X, :simplified_covering)::Covering
end
